#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Nov 27 16:46:52 2024

@author: anton
"""

from SLiCAP import *
from scipy.integrate import quad
from DIN_A import DinA

def find_gm_ciss(cir, specs, mosType):

    # Assign variables; get them from the specifications
    f_min           = cir.getParValue('f_min')
    f_max           = cir.getParValue('f_max')
    V_noise         = cir.getParValue('V_onoise')
    n_SRC           = cir.getParValue('n_SRC')
    n_Ri            = cir.getParValue('n_Ri')
    
    # Define NMOS parameters
    
    if mosType == "N":
        cir.defPar("n", "N_s_N18")
        cir.defPar("C_OX", "C_OX_N18")
        cir.defPar("KF", "KF_N18")
        cir.defPar("AF", "AF_N18")
        
    elif mosType == "P":
        cir.defPar("n", "N_s_P18")
        cir.defPar("C_OX", "C_OX_P18")
        cir.defPar("KF", "KF_P18")
        cir.defPar("AF", "AF_P18")
    
    # Assign a budget for the relative noise contribution of the MOS input stage
    n_M = 0
    while n_M > 1-n_SRC - n_Ri or n_M <= 0:
        try:
            n_M = float(input("\nAssign a relative budget for the contribution " +
                              "of the noise of the controller %sMOS input "%(mosType) +
                              "stage to the squared DIN-A weighted RMS output " +
                              "noise (> 0 and < %s):\n>>> "%(sp.N(1 - n_SRC - n_Ri , 3))))
        except ValueError:
            n_M=0
            
    N_var_required  = V_noise**2 * (n_M + n_SRC + n_Ri)
    
    # Calculate the output noise as a function of c_iss and g_m
    noiseResult = doNoise(cir, pardefs="circuit")
    S_onoise    = noiseResult.onoise
    onoiseVarA  = rmsNoise(noiseResult, 'onoise', f_min, f_max)**2
    
    ###############################################################################
    # Find absolute show stopper of the unweighted RMS output noise.
    # this depends on the process cut-off frequency
    g_m   = sp.Symbol("g_m")
    c_iss = sp.Symbol("c_iss")
    f_T   = sp.Symbol("f_T")
    onoiseVar_fT = onoiseVarA.subs(g_m, 2*sp.pi*f_T*c_iss)
    
    # At peak fT
    onoiseVar_fTpeak = onoiseVar_fT.subs(f_T, 50E9)
    
    # Minimum for c_iss
    c_iss_best = sp.solve(sp.diff(onoiseVar_fTpeak, c_iss))[1]
    # Unweighted minimum RMS noise possible:
    onoiseRMS_min = sp.sqrt(onoiseVar_fTpeak.subs(c_iss, c_iss_best))
    ###############################################################################
    
    DINA_correction = 1.
    N_var      = DINA_correction * N_var_required
    i = 0
    while i <= 1:
        onoiseVar   = assumePosParams(onoiseVarA)
        
        # we obtain valid combinations of g_m and c_iss by solving g_m from onoiseVar = N_var
        
        g_m     = sp.Symbol('g_m', positive = True)
        gm_ciss = sp.solve(onoiseVar - N_var, g_m)[0]
        
        # for positive values of both g_m and c_iss, we need the denominator to be positive
        # this requires c_iss to be larger than its solution from the denominator
        
        num, den = gm_ciss.as_numer_denom()
        ciss_min_den = sp.solve(den)[0]
    
        try:
            ciss_min_num = sp.solve(num)[0]
            ciss_min = np.max([ciss_min_num, ciss_min_den])
        except IndexError:
            ciss_min = ciss_min_den
    
        # a minimum value of g_m is found from solving the derivative of gm_ciss for
        # c_iss > ciss_min; we will do this numerically
        c_iss             = sp.Symbol('c_iss', positive = True)
        diff_gm_ciss      = sp.diff(gm_ciss, c_iss)
        func_diff_gm_ciss = sp.lambdify(c_iss, diff_gm_ciss)
        ciss_gm_min       = fsolve(func_diff_gm_ciss, float(ciss_min) + 1e-15)[0]
        gm_min            = gm_ciss.subs(c_iss, ciss_gm_min)
    
        # The product of C_iss and g_m can be taken as a measure for the costs
        # We obtain minimum cost if d(g_m*c_iss)/d(c_iss) = 0
    
        costs                = gm_ciss * c_iss
        diff_costs_ciss      = sp.diff(costs, c_iss)
        func_diff_costs_ciss = sp.lambdify(c_iss, diff_costs_ciss)
        ciss_costs_min       = fsolve(func_diff_costs_ciss, float(ciss_min) + 1e-15)[0]
        gm_costs_min         = gm_ciss.subs(c_iss, ciss_costs_min)
    
        #print("Minimum costs require g_m =", sp.N(gm_min, 3))
        #print("They are obtained with c_iss =", sp.N(ciss_gm_min, 3))
    
        # Calculate the output noise variance for minimum costs.
        ciss = ciss_costs_min
        gm  = gm_costs_min
        N_var_flat = onoiseVar.subs({c_iss: ciss, g_m: gm})
    
        # Plot the noise spectrum
        c_iss = clearAssumptions(c_iss)
        g_m   = clearAssumptions(g_m)
        noiseResultA = noiseResult.onoise.subs({c_iss: ciss, g_m: gm})
        f = ini.frequency
        func_OnoiseVar = sp.lambdify(f, sp.N(noiseResultA * DinA**2))
        N_var_dinA = quad(func_OnoiseVar, f_min, f_max)[0]
    
        # Update noise requirements after first run
        new_correction = N_var/N_var_dinA
        if i == 1:
            noiseResult.onoise = noiseResultA
            i += 1
        else:
            DINA_correction = new_correction
            N_var       = DINA_correction * N_var
            i += 1
    
    if mosType == "N":
        onoiseSpectrum = plotSweep("onoiseSpectrum_ciss_gm_N", 
                                   "Output noise spectral density (NMOS input stage)", 
                                   noiseResult, f_min, f_max, 200, funcType="onoise", 
                                   show=False)
        print("\nNMOS design results:\n--------------------")
    elif mosType == "P":
        onoiseSpectrum = plotSweep("onoiseSpectrum_ciss_gm_P", 
                                   "Output noise spectral density (PMOS input sage)", 
                                   noiseResult, f_min, f_max, 200, funcType="onoise", 
                                   show=False)
        print("\nPMOS design results:\n--------------------")
        
    print("RMS DIN A weighted required :", sp.N(sp.sqrt(N_var_required), 3), 'V')
    print("RMS unweighted designed     :", sp.N(sp.sqrt(N_var), 3), 'V')
    print("RMS unweighted obtained     :", sp.N(sp.sqrt(N_var_flat), 3), 'V')
    print("RMS DIN A weighted obtained :", sp.N(sp.sqrt(N_var_dinA), 3), 'V')
    print("DIN A RMS correction factor :", sp.N(sp.sqrt(N_var/N_var_dinA), 3))
    print("Selected value of c_iss MOS :", sp.N(ciss, 3), 'F')
    print("Required value of g_m MOS   :", sp.N(gm, 3), 'S')
    
    specs.append(
            specItem('n_M%s'%(mosType),
            description = 'Relative contribution controller %sMOS input stage'%(mosType)  +
            'noise to the squared DIN-A weighted RMS output noise',
            value       = n_M,
            units       = "",
            specType    = "budgets")
    )
    
    specs.append(
            specItem('c_iss%s'%(mosType),
            description = "Selected input capacitance %sMOS"%(mosType),
            value       = ciss,
            units       = "F",
            specType    = "design",
            )
    )
    
    specs.append(
            specItem('g_m%s'%(mosType),
            description = "Selected transconductance input stage %sMOS"%(mosType),
            value       = gm,
            units       = "S",
            specType    = "design",
            )
    )
    
    specs.append(
            specItem('g_m_min%s'%(mosType),
            description = "Minimum transconductance input stage %sMOS"%(mosType),
            value       = gm_min,
            units       = "S",
            specType    = "design",
            )
    )
    
    specs.append(
            specItem('c_iss_g_m_min%s'%(mosType),
            description = "Required c_issN at g_m_min%s"%(mosType),
            value       = ciss_gm_min,
            units       = "F",
            specType    = "design",
            )
    )
    
    specs.append(
            specItem('c_iss_costs_min%s'%(mosType),
            description = "Required c_iss %sMOS for minimum costs"%(mosType),
            value       = ciss_costs_min,
            units       = "F",
            specType    = "design",
            )
    )
    
    specs.append(
            specItem('g_m_costs_min%s'%(mosType),
            description = "g_m %sMOS for minimum costs"%(mosType),
            value       = gm_costs_min,
            units       = "S",
            specType    = "design",
            )
    )
    
    specs.append(
            specItem('V_onoise_gmCiss%s'%(mosType),
            description = "DIN A weighted RMS output noise with %sMOS and selected g_m and c_iss"%(mosType),
            value       = sp.sqrt(N_var_dinA),
            units       = "V",
            specType    = "achievements",
            )
    )
    specs2csv(specs, 'A1specs.csv')
    
    ###########################################################################
    # HTML design report                                                             #
    ###########################################################################
    
    htmlPage("Determination of %sMOS input stage $g_m$ and $c_{iss}$"%(mosType))
        
    head2html('Noise model for determination of $g_m$ and $c_{iss}$')
    img2html('noise_ciss_gm_N.svg', 1000)
    
    head2html("Determination of the output noise")
    text2html("The budget for the DIN A weighted RMS noise equals $%s\\,\\mu V$."%(sp.latex(sp.N(sp.sqrt(N_var_required)*1E6, 3))))
    
    head2html("Determination of the lowest possible unweighted RMS output noise")
    text2html("The output noise spectral density in V$^2$/Hz is obtained as:")
    eqn2html("S_vo", S_onoise)
    text2html("Notice that this spectrum decreases with increasing $g_m$. Hence, " +
              "increase of the inversion coefficient at a given $c_{iss}$ results in a lower noise.")
    text2html("The unweighted squared RMS output noise $V_{onoise}^2$ over the frequency range of interest is found as:")
    eqn2html("V_onoise^2", onoiseVarA)
    text2html("If we substitute $g_m=2\\pi f_T c_{iss}$, and take the maximum $f_T=f_{T_{peak}}, " +
              "we can calculate the minimum noise that can be obtained in this technology and for this application.")
    text2html("To this end we calculate the optimum value $c_{iss_{opt}}$ for " +
              "$c_{iss}$ from $\\frac{d}{d\\,c_{iss}}V_{onoise}^2 = 0$.")
    text2html("Substitution of $c_{iss_{opt}}$ in the expression for $V_{onoise}^2$ " +
              "yields the minimum unweighted noise$V_{onoise_{min}}$; we find: " +
              "$V_{onoise_{min}}=%s$. This is much lower than required."%(sp.latex(sp.N(onoiseRMS_min, 3))))
    
    head2html("Determination $g_m$ and $c_{iss}$ for the best performance-to-cost ratio")
    text2html("The budget for the DIN A weighted squared noise equals $%s\\, V^2$."%(sp.latex(sp.N(N_var_required, 3))))
    text2html("As a result of the DIN A weighting this budget for the unweighted squeared noise be factor $%s$ more, "%(sp.latex(sp.N(DINA_correction, 3))) +
              "which equals: $%s\\, V^2$."%(sp.latex(sp.N(N_var, 3))))
    text2html("This factor has been determined after a few iterations of the procedure below, " +
              "of which the results of the last iteration are shown.")
    text2html("The unweighted squared RMS output noise over the frequency range of interest was found as $V_{onoise}^2$.")
    text2html("We obtain valid combinations of $g_m$ and $c_{iss}$ by solving $g_m$ from:")
    eqn2html("V_onoise^2", N_var)
    text2html("This yields:")
    eqn2html("g_m", gm_ciss)
    text2html("A positive value for $g_m$ is obtained if both the numerator " +
              "and the denominator of this expression are positive.")
    text2html("This occurs if $c_{iss} > %s$."%(sp.latex(sp.N(ciss_min, 4))))
    text2html("A minimum transconductance is required if:")
    eqn2html("dg_m/dc_iss",0)
    text2html("This minimum $g_{mmin}$ is found at $c_{iss}=%s$:"%(sp.latex(sp.N(ciss_gm_min,3))))
    eqn2html("g_m_min", gm_min)
    text2html("The product $g_m c_{iss}$ can be taken as a measure for the $costs$. " +
              "The input capacitance for minimum $costs$ is obtained from:")
    eqn2html("dcosts/dc_iss",0)
    text2html("This yields: $c_{iss}=%s$ pF and $g_m > %s\\,\\mu $S."%(sp.latex(sp.N(ciss * 1e12,3)), sp.latex(sp.N(gm*1e6,3))))
    
    head2html("Design verification")
    text2html("The noise spectrum at the output of the receive coil amplifier is shown below.")
    fig2html(onoiseSpectrum, 600)
    text2html("The DIN A weighted RMS output noise over the frequency range of interest amounts $%s\\,\\mu$V."%(sp.latex(sp.N(sp.sqrt(N_var_dinA) * 1e6,3))))
    return specs