#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Dec 15 23:12:29 2023

@author: anton
"""

from SLiCAP import *
import sympy as sp
import numpy as np
from scipy.integrate import quad
from DIN_A import DinA

print("\n**********************************************************************")
print("* 1. Assign budget for noise of the integration resistor             *")
print("* 2. Calculate the maximum value of R_i                              *")
print("* 3. Select a value of R_i                                           *")
print("* 4. Calculate feedback network drive requirements.                  *")
print("**********************************************************************\n")
# Import the specifications
initSpecs = csv2specs('A1specs.csv')


cir = makeCircuit("kicad/noise2/noise2.kicad_sch")
# Load circuit parameters from specifications
specs2circuit(initSpecs, cir)

R_i                 = sp.Symbol("R_i")
f_min               = cir.getParValue('f_min')
f_max               = cir.getParValue('f_max')
V_noise             = cir.getParValue('V_onoise')
tau_i               = cir.getParValue('tau_i')
f_fp                = cir.getParValue('f_fp')
Vi_ADC              = cir.getParValue('Vi_ADC')
n_SRC               = cir.getParValue('n_SRC')
SRCnoise            = cir.getParValue('SRCnoise')

noiseResult2 = doNoise(cir, pardefs="circuit")
S_onoise     = noiseResult2.onoise

n_Ri = 0

while n_Ri > 1 - n_SRC or n_Ri <= 0:
    try:
        n_Ri = float(input("\nPlease assign a budget for the relative contribution of " +
                           "the noise of the integrator resistance to the "+
                           "squared DIN-A weighted RMS output noise " +
                           "(> 0 and < %s):\n>>> "%(sp.N(1-n_SRC,3))))
    except ValueError:
        n_Ri = 0

N_var_required  = V_noise**2 * (n_Ri + n_SRC)

i                   = 0
correction          = 1
while i <= 1:
    N_var_corrected = N_var_required * correction
    RMS_2_sq        = rmsNoise(noiseResult2, "onoise", f_min, f_max)**2
    try:
        Ri_max      = sp.solve(RMS_2_sq - N_var_corrected, R_i)[0]
    except IndexError:
        i = 2
        Ri_max = 0
    S_weighted      = S_onoise.subs(R_i, Ri_max) * DinA**2
    func_S_weighted = sp.lambdify(ini.frequency, S_weighted)
    N_var_obtained  = quad(func_S_weighted, f_min, f_max)[0]
    correction      = N_var_required/N_var_obtained
    i += 1

Ri = Ri_max + 1

while Ri > Ri_max:
    try:
        Ri = float(input("\nPlease assign a value to the integrator resistance R_i " +
                         "(<= %s Ohm):\n>>> "%(sp.N(Ri_max,5))))
    except ValueError:
        R_i = 0

Ci = tau_i/Ri
ZL_min       = sp.sqrt(Ri**2 + 1/(2*sp.pi*f_fp*Ci)**2)
IL_fb_max    = sp.N(Vi_ADC/ZL_min)

S_weighted      = S_onoise.subs(R_i, Ri) * DinA**2
func_S_weighted = sp.lambdify(ini.frequency, S_weighted)
S_obtained      = quad(func_S_weighted, f_min, f_max)[0]
S_Ri_rel        = (S_obtained - SRCnoise**2)/V_noise**2

print("\nVn:  (DIN A weighted RMS output noise)                     :", sp.N(sp.sqrt(S_obtained), 3))
print("n_Ri (Relative contribution of R_i to squared output noise):", sp.N(S_Ri_rel, 3))
print("R_i  (integrator resistance)                               :", sp.N(Ri, 3))
print("C_i  (integrator capacitance)                              :", sp.N(Ci, 3))
print("I_fb (max peak-peak current in RiCi)                       :", sp.N(IL_fb_max, 3))

updatedSpecs = [spec for spec in initSpecs]

updatedSpecs.append(
        specItem('n_Ri',
        description = "Relative contribution of R_i noise to the " +
                      "squared DIN-A weighted RMS output noise",
        value       = n_Ri,
        units       = "",
        specType    = "budgets",
        )
)

updatedSpecs.append(
        specItem('Ri_max',
        description = "Max. value of integrator resistor",
        value       = Ri_max,
        units       = "Omega",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('R_i',
        description = "Selected value of integrator resistor",
        value       = Ri,
        units       = "Omega",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('C_i',
        description = "Required value of integrator capacitor",
        value       = Ci,
        units       = "F",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('I_fb',
        description = "Maximum peak-peak current through feedback network",
        value       = IL_fb_max ,
        units       = "A",
        specType    = "design",
        )
)
specs2csv(updatedSpecs, 'A1specs.csv')

###############################################################################
# HTML Report
###############################################################################
htmlPage('Specifications')
head2html('Initial specifications')
text2html('The initial noise specifications are summerized in the table below.')
specs2html(initSpecs)
head2html('Updated specifications')
text2html('After finishing this design step the updated specifications are:')
specs2html(updatedSpecs)